from ..data_managers.semantic_header import ConstructedNodes
from ..database_managers.db_connection import Query


class DBManagementQueryLibrary:
    @staticmethod
    def get_all_rel_types_query() -> Query:
        # find all relations and return the distinct types

        # language=SQL
        query_str = '''
                MATCH () - [rel] - () RETURN DISTINCT type(rel) AS rel_type
            '''

        return Query(query_str=query_str,
                     template_string_parameters={},
                     parameters={})

    @staticmethod
    def get_all_node_labels_query() -> Query:
        # find all nodes and return the distinct labels

        # language=SQL
        query_str = '''
            MATCH (n) RETURN DISTINCT labels(n) AS label
        '''

        return Query(query_str=query_str,
                     template_string_parameters={},
                     parameters={})

    @staticmethod
    def get_clear_db_query(db_name) -> Query:
        # language=SQL
        query_str = '''
            CREATE OR REPLACE DATABASE $db_name
            WAIT
        '''

        return Query(query_str=query_str, database="system", template_string_parameters={"db_name": db_name})

    @staticmethod
    def get_constraint_unique_event_id_query() -> Query:
        # language=SQL
        query_str = '''
            CREATE CONSTRAINT unique_event_ids IF NOT EXISTS 
            FOR (e:Event) REQUIRE e.ID IS UNIQUE
        '''
        return Query(query_str=query_str,
                     template_string_parameters={},
                     parameters={})

    @staticmethod
    def get_constraint_unique_entity_uid_query() -> Query:
        # language=SQL
        query_str = '''
            CREATE CONSTRAINT unique_entity_ids IF NOT EXISTS 
            FOR (en:Entity) REQUIRE en.sysId IS UNIQUE
        '''
        return Query(query_str=query_str,
                     template_string_parameters={},
                     parameters={})

    @staticmethod
    def get_constraint_unique_log_id_query() -> Query:
        # language=SQL
        query_str = '''
            CREATE CONSTRAINT unique_entity_ids IF NOT EXISTS 
            FOR (l:Log) REQUIRE l.ID IS UNIQUE
        '''
        return Query(query_str=query_str,
                     template_string_parameters={},
                     parameters={})

    @staticmethod
    def get_set_sysid_index_query() -> Query:
        # language=SQL
        query_str = '''
            CREATE INDEX entity_sys_id_index IF NOT EXISTS FOR (n:Entity) ON (n.sysId)
        '''
        return Query(query_str=query_str,
                     template_string_parameters={},
                     parameters={})

    @staticmethod
    def get_node_count_query() -> Query:
        # language=SQL
        query_str = '''
                // List all node types and counts
                MATCH (n) 
                WITH n, CASE labels(n)[0]
                    WHEN 'Event' THEN 0
                    WHEN 'Entity' THEN 1
                    WHEN 'Class' THEN 2
                    WHEN 'Log' THEN 3
                    ELSE 4
                END AS sortOrder
                WITH  labels(n)[0] AS label,  count(n) AS numberOfNodes,sortOrder
                RETURN label,  numberOfNodes ORDER BY sortOrder
            '''

        return Query(query_str=query_str)

    @staticmethod
    def get_edge_count_query() -> Query:
        # language=SQL
        query_str = '''
                // List all agg rel types and counts
                MATCH () - [r] -> ()
                WHERE r.type is NOT NULL
                WITH r, CASE toUpper(r.type)
                  WHEN 'REL' THEN 0
                  WHEN 'DF' THEN 1
                  ELSE 2
                END as sortOrder
                WITH toUpper(r.type) as aggType, count(r) as aggNumberOfRelations, sortOrder
                RETURN aggType, aggNumberOfRelations ORDER BY sortOrder
            '''

        return Query(query_str=query_str)

    @staticmethod
    def get_aggregated_edge_count_query() -> Query:
        # language=SQL
        query_str = '''
                // List all rel types and counts
                MATCH () - [r] -> ()
                // WHERE r.type is  NULL
                WITH r, CASE Type(r)
                  WHEN 'CORR' THEN 0
                  WHEN 'OBSERVED' THEN 1
                  WHEN 'HAS' THEN 2
                  ELSE 3
                END as sortOrder
                WITH Type(r) as type, count(r) as numberOfRelations, sortOrder
                RETURN type, numberOfRelations ORDER BY sortOrder
            '''

        return Query(query_str=query_str)

    @staticmethod
    def get_event_log_query(entity: ConstructedNodes, additional_event_attributes) -> Query:
        query_str = '''
                MATCH (e:Event) - [:CORR] -> (n:$node_label)
                RETURN n.sysId as caseId, e.activity as activity, e.timestamp as timestamp $extra_attributes
                ORDER BY n.ID, e.timestamp
            '''

        attributes_query = ",".join(f"e.{attribute} as {attribute}" for attribute in additional_event_attributes)
        return Query(query_str=query_str,
                     template_string_parameters={
                         "node_label": entity.get_label_string(),
                         "extra_attributes": f", {attributes_query}" if len(additional_event_attributes) > 0 else ""
                     })



import os
from typing import List, Set, Optional, Dict

from .db_connection import DatabaseConnection
from ..data_managers.semantic_header import SemanticHeader
from ..data_managers.datastructures import ImportedDataStructures
from ..ekg_modules.ekg_analysis import EKGAnalysis
from ..ekg_modules.ekg_builder_semantic_header import EKGUsingSemanticHeaderBuilder
from ..ekg_modules.db_management import DBManagement
from ..ekg_modules.data_importer import Importer
from ..ekg_modules.inference_engine import InferenceEngine
from ..utilities.performance_handling import Performance

from tabulate import tabulate
import pandas as pd


# ensure to allocate enough memory to your database: dbms.memory.heap.max_size=5G advised
class EventKnowledgeGraph:
    """
    This is a Class that acts as an interface between the user and the different supported modules

    :param specification_of_data_structures: A specification of how the data sets are structured
    :type specification_of_data_structures: ImportedDataStructures
    :param semantic_header: a :class:SemanticHeader class describing the semantics of the EKG
    :type semantic_header: SemanticHeader
    :param perf: a :class:Performance to keep track of the running time of the EKG construction
    :type perf: Performance
    :param batch_size: the batch_size used when calling apoc.periodic.iterate --> reduce required memory, defaults to
    5000
    :type batch_size: int
    :param use_sample: boolean indicating whether the DB should be build using a sample as specified in the
    ImportedDataStructures, defaults to False
    :type use_sample: bool

    """

    def __init__(self, specification_of_data_structures: ImportedDataStructures, perf_path: str,
                 batch_size: int = 5000, use_sample: bool = False, use_preprocessed_files: bool = False,
                 semantic_header: SemanticHeader = None, custom_module_name=None, number_of_steps: int = None):
        # classes responsible for executing queries
        self.ekg_management = DBManagement()
        self.data_importer = Importer(data_structures=specification_of_data_structures,
                                      records=semantic_header.records,
                                      batch_size=batch_size,
                                      use_sample=use_sample, use_preprocessed_files=use_preprocessed_files)
        self.ekg_builder = EKGUsingSemanticHeaderBuilder(semantic_header=semantic_header,
                                                         batch_size=batch_size)
        self.inference_engine = InferenceEngine()
        self.ekg_analysis = EKGAnalysis()

        if custom_module_name is not None:
            self.custom_module = custom_module_name()
        else:
            self.custom_module = None

        self.semantic_header = semantic_header

        self.perf = Performance(perf_path=perf_path, number_of_steps=number_of_steps)

    # region EKG management
    """Define all queries and return their results (if required)"""

    def clear_db(self) -> None:
        """
        Pass on method to ekg_management to clear the entire database

        :return: None
        """

        self.ekg_management.clear_db()

    def set_constraints(self) -> None:
        """
        Pass on method to ekg_management to set the constraints on the model

        :return: None
        """

        self.ekg_management.set_constraints()

    def get_all_rel_types(self) -> List[str]:
        """
        Pass on method to ekg_management to get all possible rel types

        :return: Set of strings
        """
        return self.ekg_management.get_all_rel_types()

    def get_all_node_labels(self) -> Set[str]:
        """
        Pass on method to ekg_management to get all possible node labels
        :return: Set of strings
        """
        return self.ekg_management.get_all_node_labels()

    def get_statistics(self) -> List[Dict[str, str]]:
        """
        Pass on method to ekg_management to get the count of nodes per label and (aggregated) relation types
        :return:  List[Dict[str, str]]
        """
        return self.ekg_management.get_statistics()

    def print_statistics(self) -> None:
        """
        Print the statistics nicely using tabulate
        :return: None
        """
        print(tabulate(self.get_statistics()))

    def get_event_log(self, entity_type: str, additional_event_attributes: Optional[List[str]]):
        """
        Pass on ekg_management to data_imported to get an event log extracted from the EKG for a specific entity and
        return it

        :param entity_type: The type of the entity --> contract states that it exists
        :type entity_type: str
        :param additional_event_attributes: list of different attributes of event that should also be stored in the
        event log
        :type additional_event_attributes: List[str], optional
        :return: List[Dict[str, str]] #TODO check

        :raise ValueError: when the entity has not been defined
        """

        entity_object = self.semantic_header.get_entity(entity_type)
        if entity_object is None:
            raise ValueError(f"Entity {entity_type} is not defined in the semantic header")

        if additional_event_attributes is None:
            additional_event_attributes = []

        return self.ekg_management.get_event_log(entity_object, additional_event_attributes)

    def save_event_log(self, entity_type: str, additional_event_attributes: Optional[List[str]] = None) -> None:
        """
        Create an event log extracted from the EKG from a specific entity and store it as a csv file

        :param entity_type: The type of the entity
        :type entity_type: str
        :param additional_event_attributes: list of different attributes of event that should also be stored in the
        event log
        :type additional_event_attributes: List[str], optional
        :return: None
        """

        event_log = self.get_event_log(entity_type, additional_event_attributes)
        df = pd.DataFrame(event_log)

        current_file_path = os.path.dirname(__file__)

        dir_path = os.path.join(current_file_path, '../../..', '..', 'data', self.semantic_header.name, 'event_logs')
        file_path = os.path.join(dir_path, f"{entity_type}.csv")
        os.makedirs(dir_path, exist_ok=True)
        df.to_csv(file_path, index=True, index_label="idx")

    # endregion

    # region import events
    def import_data(self):
        """
        Pass on method to data_imported to import data, both event data as other data, as specified in the DS files
        :return: None

        """
        self.data_importer.import_data()

    # endregion
    
    def create_nodes(self, node_types: Optional[List[str]] = None) -> None:
        """
        Pass on method to ekg_builder to create relations between entities based on nodes as specified in the
        semantic header

        :param node_types: list of entity types that should be created based on nodes. In case of None,
        all entities based on nodes are created as specified in the semantic header
        :type node_types: List[str], optional
        :return: None

        """

        self.create_nodes_by_records(node_types)

    # region EKG builder using semantic header
    def create_nodes_by_records(self, node_types: Optional[List[str]] = None) -> None:
        """
        Pass on method to ekg_builder to create relations between entities based on nodes as specified in the
        semantic header

        :param node_types: list of entity types that should be created based on nodes. In case of None,
        all entities based on nodes are created as specified in the semantic header
        :type node_types: List[str], optional
        :return: None

        """

        self.ekg_builder.create_nodes_by_records(node_types)
        
    def create_relations(self,  relation_types: Optional[List[str]] = None) -> None:
        self.create_relations_using_record(relation_types)
        self.create_relations_using_relations(relation_types)

    def create_relations_using_record(self, relation_types: Optional[List[str]] = None) -> None:
        """
        Pass on method to ekg_builder to create relations between entities based on nodes as specified in the
        semantic header

        :param relation_types: list of relation types for which the relations should be created. In case of None,
        all relations based on relations are created as specified in the semantic header
        :type relation_types: List[str], optional
        :return: None

        """
        self.ekg_builder.create_relations_using_records(relation_types)

    def create_relations_using_relations(self, relation_types: Optional[List[str]] = None) -> None:
        """
        Pass on method to ekg_builder to create relations between entities based on relations as specified in the
        semantic header

        :param relation_types: list of relation types for which the relations should be created. In case of None,
        all relations based on relations are created for all relation types as specified in the semantic header
        :type relation_types: List[str], optional
        :return: None

        """
        self.ekg_builder.create_relations_using_relations(relation_types)

    def create_nodes_by_relations(self, node_types: Optional[List[str]] = None) -> None:
        """
        Pass on method to ekg_builder to create entities based on relations as specified in the semantic header

        :param node_types: list of entity types for which the entities based on relations should be created. In
        case of None, all entities based on relations are created as specified in the semantic header
        :type node_types: List[str], optional
        :return: None

        """
        self.ekg_builder.create_nodes_by_relations(node_types)

    def create_df_edges(self, entity_types: Optional[List[str]] = None, event_label: str = "Event") -> None:
        """
        Pass on method to ekg_builder to create directly follows edges between events for specific entities

        :param entity_types: list of entity types for which the DFs should be created. In case of None,
        DFs are created for all EntityTypes as specified in the semantic header
        :type entity_types: List[str], optional
        :return: None

        """
        self.ekg_builder.create_df_edges(entity_types, event_label)

    def merge_duplicate_df(self) -> None:
        """
        Pass on method to ekg_builder to merge parallel directly follows in between batching events

        :return: None
        """
        self.ekg_builder.merge_duplicate_df()

    def delete_parallel_dfs_derived(self) -> None:
        """
        Pass on method to ekg_builder to delete parallel directly follows that have been derived when reifying new
        entities

        :return: None
        """
        self.ekg_builder.delete_parallel_dfs_derived()

    def create_static_nodes_and_relations(self) -> None:
        """
        Pass on method to ekg_builder to create static nodes and relations
        No implementation yet

        :return: None
        """
        self.ekg_builder.create_static_nodes_and_relations()

    def match_entity_with_batch_position(self, entity_type: str, relative_position_type: str) -> None:
        """
        Pass on method to inference_engine to infer the batch position of a specific entity

        :param entity_type: The type of the entity
        :type entity_type: str
        :param relative_position_type: The type of the relative position
        :type relative_position_type: str
        :return: None

        :raise ValueError: when the entity has not been defined
        """
        entity = self.semantic_header.get_entity(entity_type)
        relative_position = self.semantic_header.get_entity(relative_position_type)
        if entity is None:
            raise ValueError(f"{entity_type} is not defined in semantic header")
        self.inference_engine.match_entity_with_batch_position(entity, relative_position)

    # rule B
    def infer_items_propagate_downwards_one_level(self, entity_type: str) -> None:
        """
        Pass on method to inference_engine to infer items while propagating downwards one levels

        :param entity_type: The type of the entity
        :type entity_type: str
        :return: None

        :raise ValueError: when the entity has not been defined
        """
        entity = self.semantic_header.get_entity(entity_type)
        if entity_type is None:
            raise ValueError(f"{entity_type} is not defined in semantic header")
        self.inference_engine.infer_items_propagate_downwards_one_level(entity)

    # rule C
    def infer_items_propagate_upwards_multiple_levels(self, entity_type: str, is_load=True) -> None:
        """
        Pass on method to inference_engine to infer items while propagating upwards multiple levels

        :param entity_type: The type of the entity
        :type entity_type: str
        :param is_load: indicating whether we are inferring upwards to load events (true) or unload events (false),
        defaults to True
        :type is_load: bool
        :return: None

        :raise ValueError: when the entity has not been defined
        """

        entity = self.semantic_header.get_entity(entity_type)
        if entity is None:
            raise ValueError(f"{entity_type} is not defined in semantic header")
        self.inference_engine.infer_items_propagate_upwards_multiple_levels(entity, is_load)

    # rule D
    def infer_items_propagate_downwards_multiple_level_w_batching(self, entity_type: str,
                                                                  relative_position_type: str, ) -> None:
        """
        Pass on method to inference_engine to infer items while propagating downwards multiple levels with batching

        :param entity_type: The type of the entity
        :type entity_type: str
        :param relative_position_type: The type of the relative position
        :type relative_position_type: str
        :return: None

        :raise ValueError: when the entity has not been defined in semantic header
        """
        entity = self.semantic_header.get_entity(entity_type)
        relative_position = self.semantic_header.get_entity(relative_position_type)
        if entity_type is None:
            raise ValueError(f"{entity_type} is not defined in semantic header")
        self.inference_engine.infer_items_propagate_downwards_multiple_level_w_batching(entity, relative_position)

    # endregion

    def create_df_process_model(self, entity_type: str):
        """
        Pass on method to ekg_analysis to create a DF process model

        :param entity_type: The type of the entity
        :type entity_type: str
        :param classifiers: the attributes used as classifiers
        :type classifiers: Optional[List[str]], defaults to None

        :return: None

        :raise ValueError: when the entity has not been defined
        """
        entity = self.semantic_header.get_entity(entity_type)
        if entity_type is None:
            raise ValueError(f"{entity_type} is not defined in semantic header")
        self.ekg_analysis.create_df_process_model(entity=entity)

    def do_custom_query(self, query_name: str, **kwargs: Optional[Dict[str, any]]) -> any:
        r"""
        Pass on the custom query to the custom module with optional key words arguments

        :param query_name: the name of the function that contains the query
        :type query_name: str
        :param kwargs: any optional key word argument
        :type kwargs: dict[str, any], optional keyword arguments
        :return: None

        :raise ValueError: when no custom module has been defined
        """
        if self.custom_module is None:
            raise ValueError("No custom module has been defined")
        return self.custom_module.do_custom_query(query_function=query_name, **kwargs)


    def save_perf(self):
        self.perf.finish()
        self.perf.save()



from string import Template
from typing import Optional, List, Dict, Any

import neo4j
from neo4j import GraphDatabase

from . import authentication
from .authentication import Credentials
from ..utilities.singleton import Singleton


class Query:
    def __init__(self, query_str: str, database: str = None, parameters: Optional[Dict[str, any]] = None,
                 template_string_parameters: Optional[Dict[str, any]] = None):
        if template_string_parameters is not None:
            self.query_string = Template(query_str).safe_substitute(template_string_parameters)
        else:
            self.query_string = query_str
        self.kwargs = parameters
        self.database = database


class DatabaseConnection(metaclass=Singleton):
    def __init__(self, uri: str, db_name: str, user: str, password: str, verbose: bool = False):
        self.db_name = db_name
        self.driver = self.start_connection(uri, user, password)
        self.verbose = verbose

    @staticmethod
    def start_connection(uri: str, user: str, password: str):
        # begin config
        # connection to Neo4J database
        driver = GraphDatabase.driver(uri, auth=(user, password), max_connection_lifetime=200)
        return driver

    def close_connection(self):
        self.driver.close()

    def exec_query(self, function, **kwargs):
        result = function(**kwargs)
        if result is None:
            return
        query = result.query_string
        kwargs = result.kwargs
        database = result.database
        if kwargs is None:
            kwargs = {}  # replace None value by an emtpy dictionary

        if "apoc.periodic.commit" in query:
            limit = kwargs["limit"]
            failed_batches = 1
            attempts = 0
            while failed_batches > 0 and attempts <= 10:
                result = self._exec_query(query, database, **kwargs)
                failed_batches = result[0]['failedBatches']
                kwargs["limit"] = int(limit / 2)
                kwargs["limit"] = max(10000, kwargs["limit"])
                attempts += 1
            if failed_batches > 0:
                raise Exception(f"Maximum attempts reached: {result[0]['batchErrors']}")

            return result
        else:

            return self._exec_query(query, database, **kwargs)

    def _exec_query(self, query: str, database: str = None, **kwargs) -> Optional[List[Dict[str, Any]]]:
        """
        Write a transaction of the query to  the server and return the result
        @param query: string, query to be executed
        @param database: string, Name of the database
        @return: The result of the query or None
        """

        def run_query(tx: neo4j.Transaction, _query: str, **_kwargs) -> Optional[List[Dict[str, Any]]]:
            """
                Run the query and return the result of the query
                @param tx: transaction class on which we can perform queries to the database
                @param _query: string
                @return: The result of the query or None if there is no result
            """
            # get the results after the query is executed
            try:
                _result = tx.run(_query, _kwargs).data()
            except Exception as inst:
                self.close_connection()
                print(inst)
            else:
                if _result is not None and _result != []:  # return the values if result is not none or empty list
                    return _result
                else:
                    return None

        if self.verbose:
            print(query)

        if database is None:
            database = self.db_name

        with self.driver.session(database=database) as session:
            result = session.write_transaction(run_query, query, **kwargs)
            return result

    @staticmethod
    def set_up_connection(
            credentials: Credentials = authentication.connections_map[authentication.Connections.LOCAL],
            verbose: bool = False):
        return DatabaseConnection(db_name=credentials.user, uri=credentials.uri, user=credentials.user,
                                  password=credentials.password, verbose=verbose)

    @staticmethod
    def set_up_connection_using_key(key=authentication.Connections.LOCAL, verbose: bool = False):
        credentials = authentication.connections_map[key]
        return DatabaseConnection.set_up_connection(credentials, verbose)



from typing import List, Set, Optional, Dict


from ..cypher_queries.db_managment_ql import DBManagementQueryLibrary as dbm_ql
from ..data_managers.semantic_header import ConstructedNodes
from ..database_managers.db_connection import DatabaseConnection
from ..utilities.performance_handling import Performance


class DBManagement:
    def __init__(self):
        self.connection = DatabaseConnection()
        self.db_name = self.connection.db_name


    @Performance.track()
    def clear_db(self):
        self.connection.exec_query(dbm_ql.get_clear_db_query, **{"db_name": self.db_name})

    @Performance.track()
    def set_constraints(self):
        # # for implementation only (not required by schema or patterns)
        # self.connection.exec_query(dbm_ql.get_constraint_unique_event_id_query)
        #
        # required by core pattern
        # self.connection.exec_query(dbm_ql.get_constraint_unique_entity_uid_query)
        #
        # self.connection.exec_query(dbm_ql.get_constraint_unique_log_id_query)

        self.connection.exec_query(dbm_ql.get_set_sysid_index_query)

    def get_all_rel_types(self) -> List[str]:
        """
        Find all possible rel types
        @return:
        """

        # execute the query and store the result
        result = self.connection.exec_query(dbm_ql.get_all_rel_types_query)
        # in case there are no rel types, the result is None
        # return in this case an emtpy list
        if result is None:
            return []
        # store the results in a list
        result = [record["rel_type"] for record in result]
        return result

    def get_all_node_labels(self) -> Set[str]:
        """
        Find all possible node labels
        @return: Set of strings
        """

        # execute the query and store the result
        result = self.connection.exec_query(dbm_ql.get_all_node_labels_query)
        # in case there are no labels, return an empty set
        if result is None:
            return set([])
        # some nodes have multiple labels, which are returned as a list of labels
        # therefore we need to flatten the result and take the set
        result = set([record for sublist in result for record in sublist["label"]])
        return result

    def get_statistics(self) -> List[Dict[str, any]]:
        def make_empty_list_if_none(_list: Optional[List[Dict[str, str]]]):
            if _list is not None:
                return _list
            else:
                return []

        node_count = self.connection.exec_query(dbm_ql.get_node_count_query)
        edge_count = self.connection.exec_query(dbm_ql.get_edge_count_query)
        agg_edge_count = self.connection.exec_query(dbm_ql.get_aggregated_edge_count_query)
        result = \
            make_empty_list_if_none(node_count) + \
            make_empty_list_if_none(edge_count) + \
            make_empty_list_if_none(agg_edge_count)
        return result

    def get_event_log(self, entity: ConstructedNodes, additional_event_attributes: List[str]):
        return self.connection.exec_query(dbm_ql.get_event_log_query,
                                          **{"entity": entity,
                                               "additional_event_attributes": additional_event_attributes})

    @Performance.track("query_function")
    def do_custom_query(self, query_function, **kwargs):
        return self.connection.exec_query(query_function, **kwargs)



import math
from typing import List

import numpy as np
from tqdm import tqdm

from ..data_managers.semantic_header import RecordConstructor
from ..database_managers.db_connection import DatabaseConnection
from ..data_managers.datastructures import ImportedDataStructures
from ..utilities.performance_handling import Performance
from ..cypher_queries.data_importer_ql import DataImporterQueryLibrary as di_ql
import pandas as pd


class Importer:
    def __init__(self, data_structures: ImportedDataStructures,
                 records: List["RecordConstructor"], batch_size: int,
                 use_sample: bool = False, use_preprocessed_files: bool = False):
        self.connection = DatabaseConnection()
        self.structures = data_structures.structures
        self.records = records

        self.batch_size = batch_size
        self.load_batch_size = 20000
        self.use_sample = use_sample
        self.use_preprocessed_files = use_preprocessed_files
        self.load_status = 0

    def update_load_status(self):
        self.connection.exec_query(di_ql.get_update_load_status_query,
                                   **{
                                       "current_load_status": self.load_status,
                                       "batch_size": self.batch_size
                                   })
        self.load_status += 1

    def import_data(self) -> None:
        for structure in self.structures:
            labels = structure.labels
            record_constructors = self._get_record_constructors_by_labels(structure=structure, labels=labels)
            file_directory = structure.file_directory
            # read in all file names that match this structure
            for file_name in structure.file_names:
                # read and import the nodes
                df_log = structure.read_data_set(file_directory, file_name, use_sample=self.use_sample,
                                                 use_preprocessed_file=self.use_preprocessed_files)
                df_log["loadStatus"] = self.load_status
                self._import_nodes_from_data(df_log=df_log, file_name=file_name,
                                             record_constructors=record_constructors)

            if structure.has_datetime_attribute():
                # once all events are imported, we convert the string timestamp to the timestamp as used in Cypher
                self._reformat_timestamps(structure=structure)
                self.update_load_status()

            self._filter_nodes(structure=structure)  # filter nodes according to the structure

            self._finalize_import()  # removes temporary properties

    @Performance.track("structure")
    def _reformat_timestamps(self, structure):
        datetime_formats = structure.get_datetime_formats()
        for attribute, datetime_format in datetime_formats.items():
            if datetime_format.is_epoch:
                self.connection.exec_query(di_ql.get_convert_epoch_to_timestamp_query,
                                           **{
                                               "attribute": attribute,
                                               "datetime_object": datetime_format,
                                               "batch_size": self.batch_size
                                           })

            self.connection.exec_query(di_ql.get_make_timestamp_date_query,
                                       **{
                                           "attribute": attribute, "datetime_object": datetime_format,
                                           "batch_size": self.batch_size,
                                           "load_status": self.load_status
                                       })

    @Performance.track("structure")
    def _filter_nodes(self, structure):
        for boolean in (True, False):
            attribute_values_pairs_filtered = structure.get_attribute_value_pairs_filtered(exclude=boolean)
            for name, values in attribute_values_pairs_filtered.items():
                self.connection.exec_query(di_ql.get_filter_events_by_property_query,
                                           **{
                                               "prop": name, "values": values, "exclude": boolean,
                                               "load_status": self.load_status
                                           })

    @Performance.track("structure")
    def _finalize_import(self):
        # finalize the import
        self.connection.exec_query(di_ql.get_finalize_import_events_query,
                                   **{
                                       "batch_size": self.batch_size
                                   })

    @Performance.track("file_name")
    def _import_nodes_from_data(self, df_log, file_name, record_constructors):
        # start with batch 0 and increment until everything is imported
        batch = 0
        print("\n")
        pbar = tqdm(total=math.ceil(len(df_log) / self.load_batch_size), position=0)

        labels_constructor = di_ql.get_label_constructors(record_constructors)

        while batch * self.load_batch_size < len(df_log):
            pbar.set_description(f"Loading data from {file_name} from batch {batch}")

            # import the events in batches, use the records of the log
            batch_without_nans = [{k: int(v) if isinstance(v, np.integer) else v for k, v in m.items()
                                   if (isinstance(v, list) and len(v) > 0) or (not pd.isna(v) and v is not None)}
                                  for m in
                                  df_log[batch * self.load_batch_size:(batch + 1) * self.load_batch_size].to_dict(
                                      orient='records')]

            self.connection.exec_query(di_ql.get_create_nodes_by_importing_batch_query,
                                       **{
                                           "batch": batch_without_nans,
                                           "labels_constructors": labels_constructor
                                       })

            pbar.update(1)
            batch += 1
        pbar.close()

    def _get_record_constructors_by_labels(self, structure, labels):
        constructors = []
        for record_constructor in self.records:
            if labels is not None:
                intersection = list(set(labels) & set(record_constructor.record_labels))
            else:
                intersection = record_constructor.record_labels
            if len(intersection) > 0:  # label of record constructor is in intersection or labels is not defined
                required = True
                if record_constructor.prevalent_record.where_condition != "":
                    required = False
                else:
                    for required_attribute in record_constructor.required_attributes:
                        if required_attribute == "index":
                            continue
                        if required_attribute not in structure.attributes or structure.attributes[
                            required_attribute].optional:
                            required = False
                constructors.append({"required": required, "record_constructor": record_constructor})
        return constructors



from ..database_managers.db_connection import DatabaseConnection
from ..utilities.performance_handling import Performance
from ..cypher_queries.analysis_ql import AnalysisQueryLibrary as analysis_ql


class EKGAnalysis:
    def __init__(self):
        self.connection = DatabaseConnection()

    @Performance.track("entity")
    def create_df_process_model(self, entity):
        self.connection.exec_query(analysis_ql.get_aggregate_df_relations_query,
                                   **{"entity": entity})



from tqdm import tqdm
import sys


# HERE BE DRAGONS
# python trickery to make sure that printing of progress bar from tqdm works fine
# print statements are overwritten as tqdm.write()
# https://stackoverflow.com/questions/36986929/redirect-print-command-in-python-script-through-tqdm-write
class DummyFile(object):
    file = None

    def __init__(self, file):
        self.file = file

    def write(self, x):
        # Avoid print() second call (useless \n)
        if len(x.rstrip()) > 0:
            tqdm.write(x, file=self.file)
            
    def flush(self) -> None:
        # NO-OP so Python doesn't complain about missing flush method.
        pass


# Make class to use __enter__ and __exit__ to avoid using _with_ to open and close context
# https://stackoverflow.com/questions/6796492/temporarily-redirect-stdout-stderr
class Nostdout:
    def __enter__(self):
        self.save_stdout = sys.stdout
        sys.stdout = DummyFile(sys.stdout)

    def __exit__(self, *args, **kwargs):
        sys.stdout = self.save_stdout



import json
from abc import ABC
from pathlib import Path
from string import Template
from typing import List, Any, Optional, Union, Dict

from dataclasses import dataclass

from ..utilities.auxiliary_functions import replace_undefined_value, create_list, get_id_attribute_from_label
import re


class Property:
    def __init__(self, attribute: str, value: str, node_name: Optional[str],
                 node_attribute: Optional[str], ref_node: Optional[str], ref_attribute: Optional[str]):
        self.attribute = attribute
        self.value = value
        self.node_name = node_name,
        self.node_attribute = node_attribute,
        self.ref_node = ref_node
        self.ref_attribute = ref_attribute

    @staticmethod
    def from_string(property_description):
        if property_description is None:
            return None
        if ":" in property_description:
            components = property_description.split(":")
        else:
            components = property_description.split("=")
        attribute = components[0]
        value = components[1]
        attribute = attribute.strip()
        value = value.strip()

        ref_node, ref_attribute, node_name, node_attribute = None, None, None, None
        if "." in value:
            components = value.split(".")
            ref_node = components[0]
            ref_attribute = components[1]

        if "." in attribute:
            components = attribute.split(".")
            node_name = components[0]
            node_attribute = components[1]

        return Property(attribute=attribute, value=value, node_name=node_name,
                        node_attribute=node_attribute, ref_node=ref_node, ref_attribute=ref_attribute)

    def get_pattern(self, is_set=False, node_name=None):
        if not is_set:
            return f"{self.attribute}: {self.value}"
        else:
            if node_name is None:
                return f"{self.attribute} = {self.value}"
            else:
                return f"{node_name}.{self.attribute} = COALESCE({node_name}.{self.attribute}, {self.value})"

    def __repr__(self):
        return self.get_pattern()


@dataclass()
class Node:
    def __init__(self, name: str, labels: List[str], required_properties: List[Property],
                 optional_properties: List[Property], where_condition: str):
        self.name = name
        self.labels = labels
        self.required_properties = required_properties
        self.optional_properties = optional_properties
        self.where_condition = where_condition

    @staticmethod
    def from_string(node_description: str) -> Optional["Node"]:
        if node_description is None:
            return None
        # we expect a node to be described in (node_name:Node_label)
        node_description = re.sub(r"[()]", "", node_description)
        node_components = node_description.split(":", 1)
        name = node_components[0]
        labels = ""
        where_condition = ""
        required_properties = []
        optional_properties = []
        if len(node_components) > 1:
            node_labels_prop_where = node_components[1]
            node_labels_prop_where = node_labels_prop_where.replace("'", "\"")
            if "WHERE" in node_labels_prop_where:
                labels = node_labels_prop_where.split(" WHERE ")[0]
                where_condition = node_labels_prop_where.split(" WHERE ")[1]
            elif "{" in node_labels_prop_where:
                labels = node_labels_prop_where.split(" {")[0]
                properties = node_labels_prop_where.split(" {")[1]
                properties = properties.replace("}", "")
                properties = properties.split(",")

                required_properties = [prop for prop in properties if "OPTIONAL" not in prop]
                optional_properties = [prop.replace("OPTIONAL", "") for prop in properties if "OPTIONAL" in prop]

                required_properties = [Property.from_string(prop) for prop in required_properties]
                optional_properties = [Property.from_string(prop) for prop in optional_properties]
            else:
                labels = node_labels_prop_where

        labels = labels.strip()
        labels = labels.split(":")

        return Node(name=name, labels=labels, required_properties=required_properties,
                    optional_properties=optional_properties,
                    where_condition=where_condition)

    def get_name(self, with_brackets=False):
        if with_brackets:
            return f"({self.name})"
        else:
            return self.name

    def get_condition_string(self, with_brackets=False, with_where=False, with_optional=False):
        properties = self.required_properties + self.optional_properties if with_optional else self.required_properties
        if len(properties) > 0:
            return self._get_property_string(with_brackets, with_optional)
        elif self.where_condition != "":
            return self._get_where_condition_string(with_where)
        else:
            return ""

    def _get_property_string(self, with_brackets=False, with_optional=False):
        properties = [req_prop.get_pattern() for req_prop in self.required_properties]
        if with_optional:
            properties += [f"OPTIONAL {prop.get_pattern()}" for prop in self.optional_properties]

        properties = ", ".join(properties)
        if with_brackets:
            property_string = "{$properties}"
            properties = Template(property_string).substitute(properties=properties)
        return properties

    def _get_where_condition_string(self, with_where=False):
        condition = self.where_condition
        if with_where:
            condition_string = "WHERE $where_condition"
            condition = Template(condition_string).substitute(where_condition=condition)
        return condition

    def get_pattern(self, name: Optional[str] = None, with_brackets=False, with_properties=True, forbidden_label=None):

        node_pattern_str = "$node_name"
        sep = ":"
        if self.get_label_str() != "":
            node_pattern_str = "$node_name:$node_label"
            if forbidden_label is not None:
                sep = "&"
                node_pattern_str+= "&!$forbidden_label"

        if name is None:
            node_pattern = Template(node_pattern_str).substitute(node_name=self.name,
                                                                 node_label=self.get_label_str(sep=sep),
                                                                 forbidden_label=forbidden_label)
        else:
            node_pattern = Template(node_pattern_str).substitute(node_name=name,
                                                                 node_label=self.get_label_str(sep=sep),
                                                                 forbidden_label=forbidden_label)
        if with_properties:
            node_pattern_str = "$node_pattern $condition_string"
            node_pattern = Template(node_pattern_str).substitute(node_pattern=node_pattern,
                                                                 condition_string=self.get_condition_string(
                                                                     with_brackets=True, with_where=True))
        if with_brackets:
            node_pattern_str = "($node_pattern)"
            node_pattern = Template(node_pattern_str).substitute(node_pattern=node_pattern)

        return node_pattern

    def get_label_str(self, include_first_colon=False, as_list=False, sep=":"):
        if as_list:
            str = ",".join([f'"{label}"' for label in self.labels])
            return f'[{str}]'

        if len(self.labels) > 0:
            return sep * include_first_colon + sep.join(self.labels)
        return ""

    def __repr__(self):
        return self.get_pattern(with_brackets=True)

    @staticmethod
    def from_dict(obj: Any) -> Optional["Node"]:
        return Node.from_string(obj)


class Relationship:
    def __init__(self, relation_name: str, relation_types: List[str], from_node: Node, to_node: Node,
                 properties: List[Property], where_condition: str, has_direction: bool):
        self.relation_name = relation_name
        self.relation_types = relation_types
        self.from_node = from_node
        self.to_node = to_node
        self.properties = properties
        self.where_condition = where_condition
        self.has_direction = has_direction

    def get_relation_type(self):
        return self.relation_types[0]

    def get_relation_types_str(self, include_first_colon=False, as_list=False):
        if as_list:
            str = ",".join([f'"{label}"' for label in self.relation_types])
            return f'[{str}]'
        if len(self.relation_types) > 0:
            return ":" * include_first_colon + ":".join(self.relation_types)
        return ""

    @staticmethod
    def from_string(relation_description: str) -> Optional["Relationship"]:
        if relation_description is None:
            return None

        # we expect a node to be described in (node_name:Node_label)
        relation_directions = {
            "left-to-right": {"has_direction": True, "from_node": 0, "to_node": 1},
            "right-to-left": {"has_direction": True, "from_node": 1, "to_node": 0},
            "undefined": {"has_direction": False, "from_node": 0, "to_node": 1}
        }

        nodes = re.findall(r'\([^<>]*\)', relation_description)
        _relation_string = re.findall(r'\[[^<>]*]', relation_description)[0]
        _relation_string = re.sub(r"[\[\]]", "", _relation_string)
        _relation_components = _relation_string.split(":", maxsplit=1)
        _relation_name = _relation_components[0]
        _relation_types = _relation_components[1]
        _relation_types = _relation_types.split(":")

        if ">" in relation_description:
            direction = "left-to-right"
        elif "<" in relation_description:
            direction = "right-to-left"
        else:
            direction = "undefined"

        # TODO, implement properties and where condition

        _has_direction = relation_directions[direction]["has_direction"]
        _from_node = Node.from_string(nodes[relation_directions[direction]["from_node"]])
        _to_node = Node.from_string(nodes[relation_directions[direction]["to_node"]])

        return Relationship(relation_name=_relation_name, relation_types=_relation_types,
                            from_node=_from_node, to_node=_to_node, properties=[], where_condition="",
                            has_direction=_has_direction)

    @staticmethod
    def from_dict(obj: Any) -> Optional["Relationship"]:
        return Relationship.from_string(obj)

    def get_pattern(self, name: Optional[str] = None, exclude_nodes=True, with_brackets=False):
        rel_pattern_str = "$rel_name"
        if self.get_relation_type() != "":
            rel_pattern_str = "$rel_name:$rel_type"

        name = name if name is not None else self.relation_name
        rel_pattern = Template(rel_pattern_str).substitute(rel_name=name,
                                                           rel_type=self.get_relation_type())

        if len(self.properties) > 0:
            properties_string = ",".join([prop.get_pattern() for prop in self.properties])
            rel_pattern_str = "$rel_pattern {$properties}"
            rel_pattern = Template(rel_pattern_str).substitute(rel_pattern=rel_pattern,
                                                               properties=properties_string)
        elif self.where_condition != "":
            rel_pattern_str = "$rel_pattern WHERE $where_condition"
            rel_pattern = Template(rel_pattern_str).substitute(rel_pattern=rel_pattern,
                                                               where_condition=self.where_condition)

        if exclude_nodes:
            if with_brackets:
                rel_pattern_str = "[$rel_pattern]"
                rel_pattern = Template(rel_pattern_str).substitute(rel_pattern=rel_pattern)
        else:
            from_node_pattern = self.from_node.get_pattern()
            to_node_pattern = self.to_node.get_pattern()
            rel_pattern_str = "($from_node) - [$rel_pattern] -> ($to_node)" if self.has_direction \
                else "($from_node) - [$rel_pattern] - ($to_node)"
            rel_pattern = Template(rel_pattern_str).substitute(from_node=from_node_pattern,
                                                               to_node=to_node_pattern,
                                                               rel_pattern=rel_pattern)

        return rel_pattern

    def __repr__(self):
        return self.get_pattern(exclude_nodes=False)


@dataclass
class RelationConstructorByNodes(ABC):
    from_node: Node
    to_node: Node
    prevalent_record: Union["Node", "Relationship"]

    @staticmethod
    def from_dict(obj: Any) -> Optional["RelationConstructorByNodes"]:
        if obj is None:
            return None

        _prevalent_record = RelationshipOrNode.from_string(obj.get("prevalent_record"))
        _from_node = Node.from_string(obj.get("from_node"))
        _to_node = Node.from_string(obj.get("to_node"))

        return RelationConstructorByNodes(prevalent_record=_prevalent_record, from_node=_from_node, to_node=_to_node)


class RelationshipOrNode(ABC):
    @staticmethod
    def from_string(relation_description: str) -> Union["Relationship", "Node"]:
        if relation_description is None:
            return None
        if "-" in relation_description:
            return Relationship.from_string(relation_description)
        else:
            return Node.from_string(relation_description)


@dataclass
class RelationConstructorByRelations(ABC):
    antecedents: List[Relationship]
    consequent: Relationship
    from_node: Node
    to_node: Node

    @staticmethod
    def from_dict(obj: Any) -> \
            Optional["RelationConstructorByRelations"]:
        if obj is None:
            return None

        _antecedents = [RelationshipOrNode.from_string(y) for y in obj.get("antecedents")]
        _consequent = Relationship.from_string(obj.get("consequent"))

        _from_node = _consequent.from_node
        _to_node = _consequent.to_node

        return RelationConstructorByRelations(antecedents=_antecedents, consequent=_consequent,
                                              from_node=_from_node, to_node=_to_node)

    def get_from_node_name(self):
        return self.consequent.from_node.name

    def get_to_node_name(self):
        return self.consequent.to_node.name

    def get_from_node_labels(self):
        return self.consequent.from_node.labels

    def get_to_node_labels(self):
        return self.consequent.to_node.labels

    def get_id_attribute_from_from_node(self):
        return get_id_attribute_from_label(self.from_node.labels[-1])

    def get_id_attribute_from_to_node(self):
        return get_id_attribute_from_label(self.to_node.labels[-1])

    def get_antecedent_query(self):
        if not all(isinstance(x, (Relationship, Node)) for x in self.antecedents):
            raise TypeError("Antecedents are not of type Relationship or Node")

        antecedents_query = [f"MATCH {antecedent.get_pattern()}" for antecedent in self.antecedents]
        antecedents_query = "\n".join(antecedents_query)

        return antecedents_query


@dataclass
class RelationConstructorByQuery(ABC):
    query: str

    @staticmethod
    def from_dict(obj: Any) -> Optional["RelationConstructorByQuery"]:
        if obj is None:
            return None

        _query = obj.get("query")

        return RelationConstructorByQuery(query=_query)


@dataclass
class NodesConstructorByQuery:
    query: str

    @staticmethod
    def from_dict(obj: Any) -> Optional["NodesConstructorByQuery"]:
        if obj is None:
            return None

        _query = obj.get("query")

        return NodesConstructorByQuery(query=_query)


class InferredRelationship:
    def __init__(self, record_labels: List[str] = None, relation_type: str = "CORR", event: Node = None):
        self.record_labels = record_labels if record_labels is not None else ["EventRecord"]
        self.relation_type = relation_type
        if event is None:
            event = Node.from_string("(event:Event)")
        self.event = event

    @staticmethod
    def from_dict(obj):
        if obj is None:
            return None

        _event = Node.from_string(obj.get("event"))
        _record_labels = obj.get("record_labels").split(":")
        _relation_type = obj.get("relation_type")

        return InferredRelationship(event=_event, record_labels=_record_labels, relation_type=_relation_type)

    def get_labels_str(self):
        return ":".join(self.record_labels)


class NodeConstructor:
    def __init__(self, prevalent_record: Optional[Union["Relationship", "Node"]],
                 node: Optional["Node"],
                 relation: Optional["Relationship"],
                 result: "Node",
                 set_labels: str,
                 infer_observed: bool = False,
                 infer_corr_from_event_record: bool = False,
                 infer_corr_from_entity_record: bool = False,
                 infer_corr_from_reified_parents: bool = False,
                 inferred_relationships: List[InferredRelationship] = None,
                 event_label: str = "Event",
                 corr_type: str = "CORR",
                 infer_reified_relation: bool = False):
        self.prevalent_record = prevalent_record
        self.relation = relation
        self.node = node
        self.result = result
        self.set_labels = set_labels
        self.infer_prevalence_record = prevalent_record is not None
        self.infer_observed = infer_observed
        self.infer_corr_from_event_record = infer_corr_from_event_record
        self.infer_corr_from_entity_record = infer_corr_from_entity_record
        self.infer_corr_from_reified_parents = infer_corr_from_reified_parents
        self.inferred_relationships = inferred_relationships
        self.event_label = event_label
        self.corr_type = corr_type
        self.infer_reified_relation = infer_reified_relation

    @staticmethod
    def from_dict(obj: Any) -> "NodeConstructor":
        _prevalent_record = RelationshipOrNode.from_string(obj.get("prevalent_record"))
        _node = Relationship.from_string(obj.get("node"))
        _relation = Relationship.from_string(obj.get("relation"))
        _result = Node.from_string(obj.get("result"))
        _set_labels = obj.get("set_labels")
        _infer_observed = replace_undefined_value(obj.get("infer_observed"), False)
        _infer_corr_from_event_record = replace_undefined_value(obj.get("infer_corr_from_event_record"), False)
        _infer_corr_from_entity_record = replace_undefined_value(obj.get("infer_corr_from_entity_record"), False)
        _infer_corr_from_reified_parents = replace_undefined_value(obj.get("infer_corr_from_reified_parents"), False)
        _corr_type = replace_undefined_value(obj.get("corr_type"), "CORR")
        _event_label = replace_undefined_value(obj.get("event_label"), "Event")
        _infer_reified_relation = replace_undefined_value(obj.get("infer_reified_relation"), False)

        _inferred_relations = create_list(InferredRelationship, obj.get("inferred_relationships"))

        return NodeConstructor(prevalent_record=_prevalent_record,
                               relation=_relation,
                               node=_node,
                               result=_result,
                               infer_observed=_infer_observed,
                               infer_corr_from_event_record=_infer_corr_from_event_record,
                               infer_corr_from_entity_record=_infer_corr_from_entity_record,
                               infer_corr_from_reified_parents=_infer_corr_from_reified_parents,
                               inferred_relationships=_inferred_relations,
                               corr_type=_corr_type,
                               event_label=_event_label,
                               infer_reified_relation=_infer_reified_relation,
                               set_labels=_set_labels)

    def __name__(self):
        if self.prevalent_record is not None:
            return "constructed_by_record"
        elif self.node is not None:
            return "constructed_by_node"
        elif self.relation is not None:
            return "constructed_by_relation"

    def get_label_string(self):
        return self.result.get_label_str()

    def get_labels(self, as_str=True):
        if as_str:
            return ",".join([f'"{label}"' for label in self.result.labels])
        return self.result.labels

    def get_prevalent_record_pattern(self, node_name: str = "record", forbidden_label: str = None):
        return self.prevalent_record.get_pattern(name=node_name, forbidden_label=forbidden_label)

    def get_keys(self):
        keys = []
        for prop in self.result.required_properties:
            # only check whether the ref attribute exists, if it should exist in the ref node
            if prop.ref_node is not None:
                key = prop.ref_attribute
                keys.append(key)
        return keys

    def get_where_condition(self, node_name: str = "record", include_start_and: bool = False):
        condition_str = " AND ".join(
            [f'''{node_name}.{key} IS NOT NULL AND {node_name}.{key} <> "Unknown"''' for key
             in self.get_keys()])
        if condition_str != "":
            if include_start_and:
                return f"AND {condition_str}"
            else:
                return condition_str
        return condition_str

    def get_pattern(self, name: Optional[str] = None, with_brackets=False, with_properties=True):
        return self.result.get_pattern(name, with_brackets, with_properties)

    def __repr__(self):
        return self.result.get_pattern(with_brackets=True)

    def constructed_by_record(self):
        return self.prevalent_record is not None

    def constructed_by_node(self):
        return self.node is not None

    def constructed_by_relation(self):
        return self.relation is not None

    def get_set_result_properties_query(self):
        if len(self.result.optional_properties) == 0:
            return None
        return ",".join(
            [prop.get_pattern(is_set=True, node_name=self.result.name) for prop in self.result.optional_properties])

    def get_idt_properties_query(self, node_name="n"):
        if self.result.required_properties is None:
            return None
        return ",".join(
            [f"{node_name}.{prop.attribute} as {prop.attribute}" for prop in self.result.required_properties])

    def get_set_result_labels_query(self):
        if self.set_labels is None:
            return None
        return f"{self.result.get_name()}:{self.set_labels}"


class ConstructedNodes:
    def __init__(self, node_type: str, include: bool,
                 node_constructors: List["NodeConstructor"],
                 infer_df: bool,
                 include_label_in_df: bool,
                 merge_duplicate_df: bool,
                 delete_parallel_df: bool):
        self.node_type = node_type
        self.include = include
        self.node_constructors = node_constructors
        self.infer_df = infer_df
        self.include_label_in_df = include_label_in_df
        self.merge_duplicate_df = merge_duplicate_df
        self.delete_parallel_df = delete_parallel_df

    def __repr__(self):
        return f"(:{self.get_label_string()})"

    @staticmethod
    def from_dict(obj: Any) -> Optional["ConstructedNodes"]:
        if obj is None:
            return None
        _include = replace_undefined_value(obj.get("include"), True)
        if not _include:
            return None

        _type = obj.get("type")
        node_constructors = create_list(NodeConstructor, obj.get("constructor"))
        _infer_df = replace_undefined_value(obj.get("infer_df"), False)
        _include_label_in_df = _infer_df and replace_undefined_value(obj.get("include_label_in_df"), False)
        _merge_duplicate_df = _infer_df and replace_undefined_value(obj.get("merge_duplicate_df"), False)
        _delete_parallel_df = _infer_df and obj.get("delete_parallel_df")

        constructed_node = ConstructedNodes(include=_include, node_constructors=node_constructors,
                                            node_type=_type,
                                            infer_df=_infer_df,
                                            include_label_in_df=_include_label_in_df,
                                            merge_duplicate_df=_merge_duplicate_df,
                                            delete_parallel_df=_delete_parallel_df)

        # TODO check whether names match
        return constructed_node

    def get_label_string(self):
        if len(self.node_constructors) == 0:
            return self.node_type
        return self.node_constructors[0].get_label_string()

    def get_labels(self):
        if len(self.node_constructors) == 0:
            return self.node_type
        return self.node_constructors[0].get_labels()

    def get_corr_type_strings(self):
        corr_types = list(set([node_constructor.corr_type for node_constructor in self.node_constructors]))
        return "|".join(corr_types)

    def get_df_label(self):
        if self.include_label_in_df:
            return f'DF_{self.node_type.upper()}'
        else:
            return f'DF'


class RelationConstructor:
    def __init__(self, prevalent_record: Optional[Union["Relationship", "Node"]],
                 nodes: List["Node"],
                 relations: List["Relationship"],
                 from_node: "Node",
                 to_node: "Node",
                 result: "Relationship",
                 optional_properties: List[Property],
                 model_as_node: bool,
                 infer_corr_from_reified_parents: bool):
        self.prevalent_record = prevalent_record
        self.from_node = from_node
        self.to_node = to_node
        self.relations = relations
        self.nodes = nodes
        self.result = result
        self.optional_properties = optional_properties
        self.model_as_node = model_as_node
        self.infer_corr_from_reified_parents = infer_corr_from_reified_parents

    @staticmethod
    def from_dict(obj: Any) -> "RelationConstructor":
        _prevalent_record = RelationshipOrNode.from_string(obj.get("prevalent_record"))
        _nodes = create_list(Node, obj.get("nodes"))
        _relations = create_list(Relationship, obj.get("relations"))
        _from_node = Node.from_string(obj.get("from_node"))
        _to_node = Node.from_string(obj.get("to_node"))
        _result = Relationship.from_string(obj.get("result"))
        _optional_properties = obj.get("set_optional_properties")
        if _optional_properties is not None:
            _optional_properties = _optional_properties.split(",")
            _optional_properties = [Property.from_string(prop) for prop in _optional_properties]

        _model_as_node = replace_undefined_value(obj.get("model_as_node"), False)
        _infer_corr_from_reified_parents = replace_undefined_value(obj.get("infer_corr_from_reified_parents"), False)

        return RelationConstructor(prevalent_record=_prevalent_record,
                                   relations=_relations,
                                   nodes=_nodes,
                                   from_node=_from_node,
                                   to_node=_to_node,
                                   result=_result,
                                   optional_properties=_optional_properties,
                                   model_as_node=_model_as_node,
                                   infer_corr_from_reified_parents=_infer_corr_from_reified_parents)

    def __name__(self):
        if self.prevalent_record is not None:
            return "constructed_by_record"
        elif self.nodes is not None:
            return "constructed_by_nodes"
        elif self.relations is not None:
            return "constructed_by_relations"

    def get_type(self):
        return self.result.get_relation_type()

    def get_prevalent_record_pattern(self, node_name: str):
        return self.prevalent_record.get_pattern(node_name)

    def get_keys(self):
        keys = []
        for prop in self.result.properties:
            key = prop.ref_attribute
            keys.append(key)
        return keys

    def get_where_condition(self, node_name: str = "record"):
        return " AND ".join(
            [f'''{node_name}.{key} IS NOT NULL AND {node_name}.{key} <> "Unknown"''' for key
             in self.get_keys()])

    def get_pattern(self, name: Optional[str] = None, with_brackets=False, with_properties=True):
        return self.result.get_pattern(name, with_brackets, with_properties)

    def __repr__(self):
        return self.result.get_pattern(with_brackets=False, exclude_nodes=False)

    def constructed_by_record(self):
        return self.prevalent_record is not None

    def constructed_by_nodes(self):
        return len(self.nodes) > 0

    def constructed_by_relations(self):
        return len(self.relations) > 0

    def get_set_result_properties_query(self):
        if self.optional_properties is None:
            return None
        return ",".join([prop.get_pattern(is_set=True) for prop in self.optional_properties])

    def get_relations_query(self):
        relation_queries = [f"MATCH {relation.get_pattern(exclude_nodes=False, with_brackets=True)}" for relation in
                            self.relations]
        relation_query = "\n".join(relation_queries)
        return relation_query


@dataclass
class ConstructedRelation:
    include: bool
    type: str
    relation_constructors: List["RelationConstructor"]

    @staticmethod
    def from_dict(obj: Any) -> Optional["ConstructedRelation"]:
        if obj is None:
            return None
        _include = replace_undefined_value(obj.get("include"), True)
        if not _include:
            return None

        _type = obj.get("type")
        _relation_constructors = create_list(RelationConstructor, obj.get("constructor"))
        return ConstructedRelation(_include, _type, relation_constructors=_relation_constructors)

    def __repr__(self):
        return f"[:{self.type}]"


class RecordConstructor:
    def __init__(self, record_labels: List[str],
                 required_attributes: List[str], optional_attributes: List[str], node_name: str = "record",
                 prevalent_record: Optional["Node"] = Node.from_string("(record:Record)")):
        self.node_name = node_name
        self.prevalent_record = prevalent_record
        self.record_labels = record_labels
        self.required_attributes = required_attributes
        self.optional_attributes = optional_attributes

    @staticmethod
    def from_dict(obj: Any) -> "RecordConstructor":
        if isinstance(obj, str):
            return RecordConstructor.from_str(obj)
        _prevalent_record = Node.from_string(obj.get("prevalent_record"))
        _record_labels = obj.get("record_labels").split(":")
        _required_attributes = obj.get("required_attributes")
        _optional_attributes = obj.get("optional_attributes")

        return RecordConstructor(prevalent_record=_prevalent_record, record_labels=_record_labels,
                                 required_attributes=_required_attributes,
                                 optional_attributes=_optional_attributes)

    @staticmethod
    def from_str(obj: str) -> "RecordConstructor":
        obj = re.sub("[\(\)]", "", obj)
        _node_name = obj.split(":", maxsplit=1)[0].strip()
        if _node_name == "":
            _node_name = "record"
        record_description = obj.split(":", maxsplit=1)[1].strip()
        if "WHERE" in record_description.upper():
            labels = record_description.split("WHERE")[0].replace(" ", "")
            condition_and_properties = record_description.split("WHERE")[1].strip()
            condition = condition_and_properties.split("{")[0].strip()
            properties = condition_and_properties.split("{")[1].replace("}", "").strip()
        else:
            labels = record_description.split("{")[0].replace(" ", "")
            condition = None
            properties = record_description.split("{")[1].replace("}", "").strip()

        _record_labels = labels.split(":")
        prevalent_record_str = f"({_node_name}:Record)"
        if condition is not None:
            prevalent_record_str = f"({_node_name}:Record WHERE {condition})"
        _prevalent_record = Node.from_string(prevalent_record_str)
        properties = properties.split(",")
        _required_attributes = [prop.strip() for prop in properties if "OPTIONAL" not in prop.upper()]
        _optional_attributes = [prop.replace("OPTIONAL", "").strip() for prop in properties if
                                "OPTIONAL" in prop.upper()]

        return RecordConstructor(node_name=_node_name, record_labels=_record_labels, prevalent_record=_prevalent_record,
                                 required_attributes=_required_attributes, optional_attributes=_optional_attributes)

    def get_prevalent_record_pattern(self, record_name: str = "record"):
        return self.prevalent_record.get_pattern(record_name)

    def get_additional_conditions(self, record_name: str = "record"):
        cond = self.prevalent_record.get_condition_string(with_brackets=False, with_where=False)
        if cond != "":
            return f"AND {self.prevalent_record.get_condition_string(with_brackets=False, with_where=False)}"
        return ""

    def get_required_attributes_is_not_null_pattern(self, record_name: str = "record"):
        return " AND ".join(
            [f'''{record_name}.{attribute} IS NOT NULL''' for attribute in self.required_attributes])

    def get_record_labels_pattern(self):
        return ":".join(self.record_labels)

    def get_label_list(self, as_str=True):
        if as_str:
            return "[" + ",".join([f'"{label}"' for label in self.record_labels]) + "]"
        return self.record_labels


class SemanticHeader:
    def __init__(self, name: str, version: str,
                 records: List["RecordConstructor"],
                 nodes: List[ConstructedNodes],
                 relations: List[ConstructedRelation]):
        self.name = name
        self.version = version
        self.records = records
        self.nodes = nodes
        self.relations = relations

    def get_entity(self, entity_type) -> Optional[ConstructedNodes]:
        for entity in self.nodes:
            if entity_type == entity.node_type:
                return entity
        return None

    @staticmethod
    def from_dict(obj: Any) -> Optional["SemanticHeader"]:
        if obj is None:
            return None
        _name = obj.get("name")
        _version = obj.get("version")
        _nodes = create_list(ConstructedNodes, obj.get("nodes"))
        _relations = create_list(ConstructedRelation, obj.get("relations"))
        _records = create_list(RecordConstructor, obj.get("records"))
        return SemanticHeader(name=_name, version=_version, records=_records, nodes=_nodes, relations=_relations)

    @staticmethod
    def create_semantic_header(path: Path):
        with open(path) as f:
            json_semantic_header = json.load(f)

        semantic_header = SemanticHeader.from_dict(json_semantic_header)
        return semantic_header

    def get_node_by_record_constructors(self, node_types: Optional[List[str]]) -> List[NodeConstructor]:
        node_constructors = []
        for node in self.nodes:
            if node_types is None or node.node_type in node_types:
                node_constructors.extend([
                    node_constructor for node_constructor in node.node_constructors if
                    node_constructor.constructed_by_record()
                ])
        return node_constructors

    def get_nodes_constructed_by_relations(self, node_types: Optional[List[str]] = None,
                                           only_include_delete_parallel_df=False) -> Dict[
        str, List[NodeConstructor]]:
        node_constructors = {}
        for node in self.nodes:
            if only_include_delete_parallel_df and not node.delete_parallel_df:
                continue
            if node_types is None or node.node_type in node_types:
                for node_constructor in node.node_constructors:
                    if node_constructor.constructed_by_relation():
                        if node.node_type not in node_constructors:
                            node_constructors[node.node_type] = []
                        node_constructors[node.node_type].append(node_constructor)
        return node_constructors

    def get_entities_constructed_by_query(self):
        return [entity for entity in self.nodes if
                entity.constructor_type == "EntityConstructorByQuery"]

    def get_relations_constructed_by_nodes(self, rel_types: Optional[List[str]]):
        return self._get_relations_constructed_by_specific_constructor(rel_types, constructor="nodes")

    def get_relations_constructed_by_relations(self, rel_types: Optional[List[str]]):
        return self._get_relations_constructed_by_specific_constructor(rel_types, constructor="relations")

    def get_relations_constructed_by_record(self, rel_types: Optional[List[str]]):
        return self._get_relations_constructed_by_specific_constructor(rel_types, constructor="record")

    def _get_relations_constructed_by_specific_constructor(self, rel_types, constructor):
        rel_constructors = []
        for relation in self.relations:
            if rel_types is None or relation.type in rel_types:
                if constructor == "nodes":
                    rel_constructors.extend(
                        [rel_constructor for rel_constructor in relation.relation_constructors if
                         rel_constructor.constructed_by_nodes()])
                elif constructor == "relations":
                    rel_constructors.extend(
                        [rel_constructor for rel_constructor in relation.relation_constructors if
                         rel_constructor.constructed_by_relations()])
                elif constructor == "record":
                    rel_constructors.extend(
                        [rel_constructor for rel_constructor in relation.relation_constructors if
                         rel_constructor.constructed_by_record()])
        return rel_constructors

    def get_relations_derived_from_relations(self):
        return [relation for relation in self.relations if
                "RelationConstructorByRelations" in relation.constructor_type]

    def get_relations_derived_from_query(self):
        return [relation for relation in self.relations if
                relation.constructor_type == "RelationConstructorByQuery"]



from .data_managers.semantic_header import SemanticHeader
from .database_managers.EventKnowledgeGraph import EventKnowledgeGraph, DatabaseConnection
from .database_managers import authentication
from .data_managers.datastructures import ImportedDataStructures
from .database_managers.db_connection import Query
from .utilities.performance_handling import Performance

__all__ = [SemanticHeader, EventKnowledgeGraph, DatabaseConnection, authentication, ImportedDataStructures, Query,
           Performance]
